@@ -23,10 +23,12 @@ use strict;
 use warnings;
 use autodie;
 
+use List::MoreUtils qw(none);
+
 use Lintian::Data;
 use Lintian::Tags qw(tag);
 use Lintian::Util qw(drain_pipe internal_error is_string_utf8_encoded open_gz
-  signal_number2name strip normalize_pkg_path);
+  signal_number2name strip normalize_pkg_path slurp_entire_file);
 use Lintian::SlidingWindow;
 
 use constant BLOCKSIZE => 16_384;
@@ -38,6 +40,7 @@ my $LOCALE_CODES = Lintian::Data->new('files/locale-codes', qr/\s++/);
 my $INCORRECT_LOCALE_CODES
   = Lintian::Data->new('files/incorrect-locale-codes', qr/\s++/);
 my $MULTIARCH_DIRS = Lintian::Data->new('common/multiarch-dirs', qr/\s++/);
+my $ALLOWED_PYTHON_FILES = Lintian::Data->new('files/allowed-python-files');
 my $GENERIC_HEADER_FILES = Lintian::Data->new('files/generic-header-files');
 my $GENERIC_PYTHON_MODULES= Lintian::Data->new('files/generic-python-modules');
 my $ALLOWED_ANCIENT_FILES = Lintian::Data->new('files/allowed-ancient-files');
@@ -188,6 +191,8 @@ my $BUILD_PATH_REGEX
   = Lintian::Data->new('files/build-path-regex',qr/~~~~~/,
     sub { return  qr/$_[0]/xsm;});
 
+my @ALLOWED_USES_DPKG_DATABASE = qw(base-files dpkg lintian);
+
 sub _tag_build_tree_path {
     my ($path, $msg) = @_;
     foreach my $buildpath ($BUILD_PATH_REGEX->all) {
@@ -524,10 +529,18 @@ sub run {
                 }
                 # -----------------/etc/apt/sources
                 if ($fname =~ m,^etc/apt/sources\.list(?:$|\.d/[^/]+),) {
-                    unless ($source_pkg eq 'apt') {
+                    unless ($source_pkg eq 'apt'
+                        or $pkg =~ /-apt-source$/) {
                         tag 'package-installs-apt-sources', $file;
                     }
                 }
+                # -----------------/etc/apt/trusted.gpg
+                if ($fname =~ m,^etc/apt/trusted\.gpg(?:$|\.d/[^/]+),) {
+                    unless ($source_pkg eq 'apt'
+                        or $pkg =~ /(?:-apt-source|-archive-keyring)$/) {
+                        tag 'package-installs-apt-keyring', $file;
+                    }
+                }
             }
             # ---------------- /etc/cron.daily, etc.
             elsif ($fname
@@ -911,7 +924,7 @@ sub run {
                     and $fname !~ m,^usr/bin/(?:X11|mh)/,) {
                     tag 'subdir-in-usr-bin', $file;
                 }
-                # check old style config script
+                # check old style config scripts
                 elsif ( $file->is_regular_file
                     and $fname =~ m,-config$,
                     and $script{$file}) {
@@ -1449,6 +1462,10 @@ sub run {
                 tag 'python-module-has-overly-generic-name', $fname, "($1)"
                   if $rest =~ m,^($regex)(?:\.py|/__init__\.py)$,i;
             }
+            tag 'unknown-file-in-python-module-directory', $fname
+              if $file->is_file
+              and $rest eq $file->basename  # "top-level"
+              and not $ALLOWED_PYTHON_FILES->matches_any($file->basename, 'i');
         }
 
         if ($fname =~ m,/icons/[^/]+/(\d+)x(\d+)/(?!animations/).*\.png$,){
@@ -1488,6 +1505,28 @@ sub run {
               and $source_pkg ne 'lintian'
               and detect_sensible_utils($file);
 
+            # ---------------- using dpkg internals
+            if ($fname !~ m,^usr/share/(?:doc|locale)/,
+                and none { $_ eq $source_pkg } @ALLOWED_USES_DPKG_DATABASE) {
+                my $strings = slurp_entire_file($info->strings($file));
+
+                # If we have strings(1) output (eg. we are an ELF
+                # binary) then prefer that.
+                if ($strings) {
+                    tag 'uses-dpkg-database-directly', $file
+                      if $strings =~ m,^/var/lib/dpkg,m;
+                } else {
+                    my $fd = $file->open(':raw');
+                    my $sfd = Lintian::SlidingWindow->new($fd);
+                    while (my $block = $sfd->readwindow) {
+                        next unless $block =~ m,/var/lib/dpkg,;
+                        tag 'uses-dpkg-database-directly', $file;
+                        last;
+                    }
+                    close($fd);
+                }
+            }
+
             # ---------------- documentation files
             unless($fname =~ m,^etc/, or $fname =~ m,^usr/share/doc/,) {
                 foreach my $taboo ($DOCUMENTATION_FILE_REGEX->all) {
@@ -1994,6 +2033,14 @@ sub run {
 
         tag 'missing-depends-on-sensible-utils', $file
           if not $has_sensible_utils and detect_sensible_utils($file);
+
+        my $fd2 = $file->open;
+        while (<$fd2>) {
+            tag 'uses-dpkg-database-directly', $file, "(line $.)"
+              if m,/var/lib/dpkg,
+              and none { $_ eq $source_pkg } @ALLOWED_USES_DPKG_DATABASE;
+        }
+        close($fd2);
     }
     close($fd);
 
